(* Copyright (c) 2011 - 2015. Cary Jensen, Jensen Data Systems, Inc.
   This code sample is intended for the readers of
   "Delphi in Depth: ClientDataSets"
     and
   "Delphi in Depth: ClientDataSets 2nd Edition"
   by Cary Jensen

   No guarantees or warranties are expressed or implied concerning
   the applicability of techniques or code included in this example
   or in the accompanying book. If you wish to use techniques or
   code included in this example or described in the book, it is
   your responsibility to test and certify any code, techniques,
   or design you adopt.

   For information on consulting or training services, please visit:
   http://www.JensenDataSystems.com

   For more information about "Delphi in Depth: ClientDataSets 2nd Edition"
   by Cary Jensen, including links to order the book, please visit:
   http://www.JensenDataSystems.com/cdsbook2
*)
unit mainformu;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  DB, Grids, DBGrids, StdCtrls, ExtCtrls, DBCtrls, Provider,
  DBClient, Variants,
  FireDAC.Comp.Client,
  FireDAC.Comp.DataSet,
  FireDAC.Comp.UI,
  FireDAC.DApt,
  FireDAC.DApt.Intf,
  FireDAC.DatS,
  FireDAC.Phys,
  FireDAC.Phys.IB,
  FireDAC.Phys.IBBase,
  FireDAC.Phys.Intf,
  FireDAC.Stan.Async,
  FireDAC.Stan.Def,
  FireDAC.Stan.Error,
  FireDAC.Stan.Intf,
  FireDAC.Stan.Option,
  FireDAC.Stan.Param,
  FireDAC.Stan.Pool,
  FireDAC.UI.Intf,
  {$IF CompilerVersion < 28}
  // nothing needed
  {$ELSE}
  FireDAC.Phys.IBDef,
  {$ENDIF}
  FireDAC.VCLUI.Wait;

type
  ECustomException = class(Exception);
  ECustomCannotFindRecord = class(EcustomException);
  ENoChanges = class(ECustomException);
  TForm1 = class(TForm)
    Panel1: TPanel;
    Button1: TButton;
    DBGrid1: TDBGrid;
    DataSource1: TDataSource;
    Button2: TButton;
    DBNavigator1: TDBNavigator;
    ClientDataSet1: TClientDataSet;
    DataSetProvider1: TDataSetProvider;
    FDGUIxWaitCursor1: TFDGUIxWaitCursor;
    FDQuery1: TFDQuery;
    FDConnection1: TFDConnection;
    FDTable1: TFDTable;
    ClientDataSet1CUSTNO: TFloatField;
    ClientDataSet1COMPANY: TStringField;
    ClientDataSet1ADDR1: TStringField;
    ClientDataSet1ADDR2: TStringField;
    ClientDataSet1CITY: TStringField;
    ClientDataSet1STATE: TStringField;
    ClientDataSet1ZIP: TStringField;
    ClientDataSet1COUNTRY: TStringField;
    ClientDataSet1PHONE: TStringField;
    ClientDataSet1FAX: TStringField;
    ClientDataSet1TAXRATE: TFloatField;
    ClientDataSet1CONTACT: TStringField;
    ClientDataSet1LASTINVOICEDATE: TSQLTimeStampField;
    procedure FormCreate(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure ClientDataSet1AfterDelete(DataSet: TDataSet);
    procedure DataSetProvider1BeforeUpdateRecord(Sender: TObject;
      SourceDS: TDataSet; DeltaDS: TCustomClientDataSet;
      UpdateKind: TUpdateKind; var Applied: Boolean);
    procedure DataSetProvider1BeforeApplyUpdates(Sender: TObject;
      var OwnerData: OleVariant);
    procedure DataSetProvider1AfterApplyUpdates(Sender: TObject;
      var OwnerData: OleVariant);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

procedure TForm1.FormCreate(Sender: TObject);
var
  DataDir: string;
begin
  DataDir := GetEnvironmentVariable('DEMOSDIR');
  if DataDir.IsEmpty then
  begin
    ShowMessage('DEMOSDIR environment variable is not set. ' +
                'Cannot continue. See Appendix A, Setting the ' +
                'DEMOSDIR environment variable for more information');
    exit;
  end;

  FDConnection1.Params.Clear;
  FDConnection1.Params.Add('DriverID=IB');
  FDConnection1.Params.Add('Database=' +
    GetEnvironmentVariable('DEMOSDIR') + 'Data\dbdemos.gdb');
  FDConnection1.Params.Add('User_Name=sysdba');
  FDConnection1.Params.Add('password=masterkey');
  FDConnection1.Params.Add('Server=localhost');
  FDCOnnection1.LoginPrompt := False;
  ClientDataSet1.Open;
end;

procedure TForm1.Button1Click(Sender: TObject);
begin
if ClientDataSet1.State in [dsInsert, dsEdit] then
  ClientDataSet1.Post;
if ClientDataSet1.ChangeCount > 0 then
begin
  if ClientDataSet1.ApplyUpdates(-1) = 0 then
  begin
    Button1.Enabled := False;
    Button2.Enabled := False;
  end;
end
else
  begin
    ShowMessage('No changes to apply');
    Button1.Enabled := False;
    Button2.Enabled := False;
  end;
end;

procedure TForm1.Button2Click(Sender: TObject);
begin
if ClientDataSet1.ChangeCount > 0 then
  begin
    ClientDataSet1.CancelUpdates;
    Button1.Enabled := False;
    Button2.Enabled := False;
  end
else
  ShowMessage('No changes to cancel');
end;

procedure TForm1.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
if ClientDataSet1.State in [dsInsert, dsEdit] then
  ClientDataSet1.Post;
if ClientDataSet1.ChangeCount > 0 then
  if MessageDlg('Updates pending. Select OK to apply, Cancel to lose',
     mtInformation, [mbOK,mbCancel],0) <> mrOK then
     ClientDataSet1.CancelUpdates
  else
    ClientDataSet1.ApplyUpdates(-1);
end;

procedure TForm1.ClientDataSet1AfterDelete(DataSet: TDataSet);
begin
Button1.Enabled := True;
Button2.Enabled := True;
end;

procedure TForm1.DataSetProvider1AfterApplyUpdates(Sender: TObject;
  var OwnerData: OleVariant);
begin
  FDTable1.Close;
end;

procedure TForm1.DataSetProvider1BeforeApplyUpdates(Sender: TObject;
  var OwnerData: OleVariant);
begin
  FDTable1.Open;
end;

procedure TForm1.DataSetProvider1BeforeUpdateRecord(Sender: TObject;
  SourceDS: TDataSet; DeltaDS: TCustomClientDataSet;
  UpdateKind: TUpdateKind; var Applied: Boolean);
var
  i: Integer;
begin
  // Prevent the default behavior
  Applied := True;
  // Post pending cached edits
  // Update the record
  if UpdateKind = ukInsert then
    begin
      FDTable1.Insert;
      for i := 0 to FDTable1.Fields.Count - 1 do
        if not DeltaDS.Fields[i].IsNull then
        FDTable1.Fields[i].Value := DeltaDS.Fields[i].Value;
      try
        FDTable1.Post;
      except
        FDTable1.Cancel;
        raise;
      end;
    end
  else
  // Not an insert. Locate the existing record
  if FDTable1.Locate('CustNo',DeltaDS.Fields[0].OldValue,[]) then
    case UpdateKind of
      ukModify:
        begin
        FDTable1.Edit;
        for i := 0 to FDTable1.Fields.Count - 1 do
          if DeltaDS.Fields[i].Value <> Null then
            FDTable1.Fields[i].Value := DeltaDS.Fields[i].Value;
        try
          FDTable1.Post;
        except;
          FDTable1.Cancel;
          raise;
        end;
        end;
      ukDelete:
        begin
        FDTable1.Delete;
        end;
    end
  else
    raise ECustomCannotFindRecord.Create(DeltaDS.Fields[0].AsString +
      ' no longer exists');
end;

end.
